﻿
' VB2008, Net Framework 2.0

' --------------------------------------------------------------------------------

' How to?

' Dim f As IFilter = New ConvolutionFilter(ConvolutionFilter.Blur, 0.5, 50.0)
' Dim d As New Image2D(OldImage)
' Call f.ApplyTo(d)
' Dim NewImage As Bitmap = d.ToBitmap()

' ---------------------------------------------------------------------------------------------


Imports System.Drawing
Imports System.Drawing.Imaging
Imports System.Runtime.InteropServices

Namespace Filters

#Region " Argb"

Public Structure Argb
   Public b As Byte
   Public g As Byte
   Public r As Byte
   Public a As Byte
   Public Sub New(ByRef a As Byte, ByRef r As Byte, ByRef g As Byte, ByRef b As Byte)
      Me.a = a : Me.r = r : Me.g = g : Me.b = b
   End Sub
   ' compositing 0.0 - 1.0
   Public Shared Function OverlayRgb(ByRef top As Argb, ByRef back As Argb, _
      Optional ByVal compositing As Single = 0.5) As Argb
      Dim n As New Argb(back.a, 255, 255, 255)
      If compositing >= 1.0 Then
         n = top
      ElseIf compositing <= 0.0 Then
         n = back
      Else
         If top.r > back.r Then
            n.r = back.r + ((top.r - back.r) * compositing)
         ElseIf top.r < back.r Then
            n.r = back.r - ((back.r - top.r) * compositing)
         Else
            n.r = back.r
         End If
         If top.g > back.g Then
            n.g = back.g + ((top.g - back.g) * compositing)
         ElseIf top.g < back.g Then
            n.g = back.g - ((back.g - top.g) * compositing)
         Else
            n.g = back.g
         End If
         If top.b > back.b Then
            n.b = back.b + ((top.b - back.b) * compositing)
         ElseIf top.b < back.b Then
            n.b = back.b - ((back.b - top.b) * compositing)
         Else
            n.b = back.b
         End If
      End If
      Return n
   End Function
End Structure

#End Region

#Region " Image2D"

Public Class Image2D

Implements ICloneable

Friend data() As Byte
Public width As Int32
Public height As Int32

Public Function Clone() As Object Implements System.ICloneable.Clone
   Dim i2d As New Image2D(width, height)
   i2d.data = data.Clone()
   Return i2d
End Function

Public Sub New(ByVal width As Int32, ByVal height As Int32)
   Me.width = width : Me.height = height
   ReDim data((width * height * 4) - 1)
End Sub

Public Sub New(ByVal source As Bitmap)
   MyBase.New()
   Me.width = source.Width : Me.height = source.Height
   ReDim data((width * height * 4) - 1)
   Using bm As New Bitmap(source.Width, source.Height, PixelFormat.Format32bppArgb)
      Call bm.SetResolution(source.HorizontalResolution, source.VerticalResolution)
      Using g As Graphics = Graphics.FromImage(bm)
         Call g.DrawImage(source, 0, 0)
      End Using
      Dim bmdata As BitmapData = bm.LockBits(New Rectangle(0, _
      0, width, height), ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb)
      Call Marshal.Copy(bmdata.Scan0, data, 0, data.Length)
      Call bm.UnlockBits(bmdata)
   End Using
End Sub

Public Sub SetPixel(ByRef x As Int32, ByRef y As Int32, ByRef value As Argb)
   On Error Resume Next
   Dim ofs As Int32 = ((y * width) + x) * 4
   data(ofs) = value.b
   data(ofs + 1) = value.g
   data(ofs + 2) = value.r
   data(ofs + 3) = value.a
End Sub

Public Function GetPixel(ByRef x As Int32, ByRef y As Int32) As Argb
   On Error Resume Next
   Dim ofs As Int32 = ((y * width) + x) * 4
   Return New Argb(data(ofs + 3), data(ofs + 2), data(ofs + 1), data(ofs))
End Function

Public Function ToBitmap() As Bitmap
   Dim bm As New Bitmap(width, height, PixelFormat.Format32bppArgb)
   Dim bmdata As BitmapData = bm.LockBits(New Rectangle(0, 0, _
   width, height), ImageLockMode.WriteOnly, PixelFormat.Format32bppArgb)
   Call Marshal.Copy(data, 0, bmdata.Scan0, data.Length)
   Call bm.UnlockBits(bmdata)
   Return bm
End Function

End Class

#End Region

Public Interface IFilter
   Sub ApplyTo(ByRef i2d As Image2D)
End Interface

Public Enum ConvolutionFilterPreset As Byte
   '[Default] = 0: Sharpen: Blur
   Sharpen = 0 : Blur
End Enum

#Region " ConvolutionFilter"

Public Class ConvolutionFilter

Implements IFilter

Private matrix(2, 2) As Single
Private compositing As Single

Private Function CreateMatrix(ByVal preset As ConvolutionFilterPreset, ByVal strength As Single) As Single(,)
   Select Case preset
      'Case ConvolutionFilterPreset.Default
      '     Return New Single(,) {{0, 0, 0}, _
      '                           {0, 1, 0}, _
      '                           {0, 0, 0}}
      Case ConvolutionFilterPreset.Sharpen
           Return New Single(,) {{-strength, -strength, -strength}, _
                                 {-strength, 1 + (strength * 8), -strength}, _
                                 {-strength, -strength, -strength}}
      Case ConvolutionFilterPreset.Blur
           Return New Single(,) {{strength, -strength, strength}, _
                                 {-strength, strength + 1, -strength}, _
                                 {strength, -strength, strength}}
   End Select
End Function

' strength 0.0 - 1.0
' compositing 0.0% - 100.0%
Public Sub New(ByVal preset As ConvolutionFilterPreset, _
   Optional ByVal strength As Single = 0.5, _
   Optional ByVal compositing As Single = 50.0)
   If (strength < 0.001) OrElse (strength > 1) Then strength = 0.5
   If (compositing < 1) OrElse (compositing > 100) Then compositing = 50.0
   Me.matrix = CreateMatrix(preset, strength)
   Me.compositing = compositing
End Sub

Public Sub New(ByVal matrix(,) As Single, Optional ByVal compositing As Single = 50.0)
   Me.matrix = matrix
   If (compositing < 1) OrElse (compositing > 100) Then compositing = 50.0
   Me.compositing = compositing
End Sub

Public Sub ApplyTo(ByRef i2d As Image2D) Implements IFilter.ApplyTo
   Dim clone As Image2D = i2d.Clone()
   Dim v, x, y As Integer, block(2, 2), c As Argb
   Dim s As Single = compositing / 100
   ' --------------------------------------------------------------------------
   For x = 1 To i2d.width - 2
       For y = 1 To i2d.height - 2
           c = i2d.GetPixel(x, y)
           ' -------------------------------------  
           ' 3 x 3  
           ' --------------------------------------
           block(0, 0) = i2d.GetPixel(x - 1, y - 1)
           block(0, 1) = i2d.GetPixel(x, y - 1)
           block(0, 2) = i2d.GetPixel(x + 1, y - 1)
           block(1, 0) = i2d.GetPixel(x - 1, y)
           block(1, 1) = i2d.GetPixel(x, y)
           block(1, 2) = i2d.GetPixel(x + 1, y)
           block(2, 0) = i2d.GetPixel(x - 1, y + 1)
           block(2, 1) = i2d.GetPixel(x, y + 1)
           block(2, 2) = i2d.GetPixel(x + 1, y + 1)
           ' ------------------------------------------
           ' R  
           v = matrix(0, 0) * block(0, 0).r + _
               matrix(0, 1) * block(0, 1).r + _
               matrix(0, 2) * block(0, 2).r + _
               matrix(1, 0) * block(1, 0).r + _
               matrix(1, 1) * block(1, 1).r + _
               matrix(1, 2) * block(1, 2).r + _
               matrix(2, 0) * block(2, 0).r + _
               matrix(2, 1) * block(2, 1).r + _
               matrix(2, 2) * block(2, 2).r
           v *= 1
           If v < 0 Then v = 0
           If v > 255 Then v = 255
           ' ------------------------------
           If v > c.r Then
              v = (v - c.r) * s
              c.r += v
           ElseIf v < c.r Then
              v = (c.r - v) * s
              c.r -= v
           End If
           ' ------------------------------
           ' G  
           v = matrix(0, 0) * block(0, 0).g + _
               matrix(0, 1) * block(0, 1).g + _
               matrix(0, 2) * block(0, 2).g + _
               matrix(1, 0) * block(1, 0).g + _
               matrix(1, 1) * block(1, 1).g + _
               matrix(1, 2) * block(1, 2).g + _
               matrix(2, 0) * block(2, 0).g + _
               matrix(2, 1) * block(2, 1).g + _
               matrix(2, 2) * block(2, 2).g
           v *= 1
           If v < 0 Then v = 0
           If v > 255 Then v = 255
           ' ------------------------------
           If v > c.g Then
              v = (v - c.g) * s
              c.g += v
           ElseIf v < c.g Then
              v = (c.g - v) * s
              c.g -= v
           End If
           ' ------------------------------
           ' B  
           v = matrix(0, 0) * block(0, 0).b + _
               matrix(0, 1) * block(0, 1).b + _
               matrix(0, 2) * block(0, 2).b + _
               matrix(1, 0) * block(1, 0).b + _
               matrix(1, 1) * block(1, 1).b + _
               matrix(1, 2) * block(1, 2).b + _
               matrix(2, 0) * block(2, 0).b + _
               matrix(2, 1) * block(2, 1).b + _
               matrix(2, 2) * block(2, 2).b
           v *= 1
           If v < 0 Then v = 0
           If v > 255 Then v = 255
           ' ------------------------------
           If v > c.b Then
              v = (v - c.b) * s
              c.b += v
           ElseIf v < c.b Then
              v = (c.b - v) * s
              c.b -= v
           End If
           ' ------------------------------
           Call clone.SetPixel(x, y, c)
       Next
   Next
   ' -------------------------------------------
   i2d.data = clone.data
End Sub

End Class

#End Region

Public Enum FilterOffset As Byte
   ofs1 = 0 : ofs2 : ofs3
End Enum

#Region " GaussianBlur 4 прохода"

Public Class GaussianBlur

Implements IFilter

Private offset As Int32
Private compositing As Single

' compositing 0.0 - 1.0
Public Sub New(Optional ByVal offset As FilterOffset = FilterOffset.ofs1, Optional ByVal compositing As Single = 0.5)
   MyBase.New()
   Me.offset = CInt(offset) + 1
   If (compositing < 0.01) OrElse (compositing > 1) Then compositing = 0.5
   Me.compositing = compositing
End Sub

Public Sub ApplyTo(ByRef i2d As Image2D) Implements IFilter.ApplyTo
   Dim w, h As Int32, pcur, pnext As Argb
   w = i2d.width - 1 : h = i2d.height - 1
   For y As Int32 = 0 To h - offset Step 1
       For x As Int32 = 0 To w - offset Step 1
           pcur = i2d.GetPixel(x, y)
           pnext = Argb.OverlayRgb(pcur, i2d.GetPixel(x + offset, y + offset), compositing)
           Call i2d.SetPixel(x + offset, y + offset, pnext)
       Next
   Next
   For y As Int32 = h To offset Step -1
       For x As Int32 = 0 To w - offset Step 1
           pcur = i2d.GetPixel(x, y)
           pnext = Argb.OverlayRgb(pcur, i2d.GetPixel(x + offset, y - offset), compositing)
           Call i2d.SetPixel(x + offset, y - offset, pnext)
       Next
   Next
   For y As Int32 = 0 To h - offset Step 1
       For x As Int32 = w To offset Step -1
           pcur = i2d.GetPixel(x, y)
           pnext = Argb.OverlayRgb(pcur, i2d.GetPixel(x - offset, y + offset), compositing)
           Call i2d.SetPixel(x - offset, y + offset, pnext)
       Next
   Next
   For y As Int32 = h To offset Step -1
       For x As Int32 = w To offset Step -1
           pcur = i2d.GetPixel(x, y)
           pnext = Argb.OverlayRgb(pcur, i2d.GetPixel(x - offset, y - offset), compositing)
           Call i2d.SetPixel(x - offset, y - offset, pnext)
       Next
   Next
End Sub

End Class

#End Region

#Region " MedianBlur 4 прохода (Bonustrack :D)"

Public Class MedianBlur

Implements IFilter

Private offset As Int32
Private compositing As Single
Private вязкость As Single

' compositing 0.0 - 1.0
' вязкость_материала 0.0 - 1.0
Public Sub New(Optional ByVal offset As FilterOffset = FilterOffset.ofs1, _
   Optional ByVal compositing As Single = 0.5, Optional ByVal вязкость_материала As Single = 0.25)
   MyBase.New()
   Me.offset = CInt(offset) + 1
   If (compositing < 0.01) OrElse (compositing > 1) Then compositing = 0.5
   Me.compositing = compositing
   If (вязкость_материала < 0.01) OrElse (вязкость_материала > 1) Then вязкость_материала = 0.5
   вязкость = вязкость_материала
End Sub

Public Sub ApplyTo(ByRef i2d As Image2D) Implements IFilter.ApplyTo
    Dim w, h As Int32, pcur, pmid As Argb
    w = i2d.width - 1 : h = i2d.height - 1
    pmid = i2d.GetPixel(0, 0)
    For y As Int32 = 0 To h
        For x As Int32 = 0 To w
            pcur = i2d.GetPixel(x, y)
            pmid = Argb.OverlayRgb(pcur, pmid, вязкость)
            pcur = Argb.OverlayRgb(pmid, pcur, compositing)
            Call i2d.SetPixel(x, y, pcur)
        Next
    Next
    For x As Int32 = w To 0 Step -1
        For y As Int32 = h To 0 Step -1
            pcur = i2d.GetPixel(x, y)
            pmid = Argb.OverlayRgb(pcur, pmid, вязкость)
            pcur = Argb.OverlayRgb(pmid, pcur, compositing)
            Call i2d.SetPixel(x, y, pcur)
        Next
    Next
    For y As Int32 = 0 To h
        For x As Int32 = w To 0 Step -1
            pcur = i2d.GetPixel(x, y)
            pmid = Argb.OverlayRgb(pcur, pmid, вязкость)
            pcur = Argb.OverlayRgb(pmid, pcur, compositing)
            Call i2d.SetPixel(x, y, pcur)
        Next
    Next
    For x As Int32 = 0 To w
        For y As Int32 = h To 0 Step -1
            pcur = i2d.GetPixel(x, y)
            pmid = Argb.OverlayRgb(pcur, pmid, вязкость)
            pcur = Argb.OverlayRgb(pmid, pcur, compositing)
            Call i2d.SetPixel(x, y, pcur)
        Next
    Next
End Sub

End Class

#End Region

End Namespace